import datetime
from typing import Dict, Literal, Optional, Union

import pandas as pd

from relbench.base import Database

from .db_dataset import DBDataset

__ALL__ = [
    "Accidents",
    "Airline",
    "Expenditures",
    "Employee",
    "LegalActs",
    "Seznam",
    "TPCC",
    "TPCD",
    "TPCDS",
    "TPCH",
    "VOC",
    "Walmart",
]

# fmt: off
CTUDatabaseName = Literal[
    'Accidents', 'AdventureWorks2014', 'Airline', 'Atherosclerosis', 'AustralianFootball', 
    'Basketball_men', 'Basketball_women', 'Biodegradability', 'Bupa', 'Carcinogenesis', 
    'ccs', 'CDESchools', 'Chess', 'CiteSeer', 'classicmodels', 'ConsumerExpenditures',
    'CORA', 'Countries', 'CraftBeer', 'Credit', 'cs', 'Dallas', 'DCG', 'Dunur', 'Elti',
    'employee', 'ErgastF1', 'Facebook', 'financial', 'FNHK', 'ftp', 'geneea', 'genes',
    'GOSales', 'Grants', 'Hepatitis_std', 'Hockey', 'imdb_ijs', 'KRK', 'lahman_2014',
    'legalActs', 'Mesh', 'medical', 'Mondial', 'Mooney_Family', 'imdb_MovieLens',
    'MuskSmall', 'mutagenesis', 'nations', 'NBA', 'NCAA', 'northwind', 'Pima',
    'PremierLeague', 'PTE', 'PubMed_Diabetes', 'pubs', 'Pyrimidine', 'restbase',
    'sakila', 'SalesDB', 'Same_gen', 'SAP', 'SAT', 'Seznam', 'SFScores', 'Shakespeare',
    'stats', 'Student_loan', 'Toxicology', 'tpcc', 'tpcd', 'tpcds', 'tpch', 'trains',
    'Triazine', 'university', 'UTube', 'UW_std', 'VisualGenome', 'voc', 'Walmart','WebKP',
    'world'
]
# fmt: on


class CTUDataset(DBDataset):
    def __init__(
        self,
        database: CTUDatabaseName,
        cache_dir: Optional[str] = None,
        time_col_dict: Optional[Dict[str, str]] = None,
        keep_original_keys: bool = False,
    ):
        """Create a database dataset object.

        Args:
            database (CTUDatabaseName): The name of the database.
            cache_dir (str, optional): The directory to cache the dataset. Defaults to None.
            time_col_dict (Dict[str, str], optional): A dictionary mapping table names to time columns. Defaults to None.
            keep_original_keys (bool, optional): Whether to keep original primary and foreign keys \
                after duplication during re-indexing. This is useful when the keys contain information \
                beyond just their relationship to other rows. Defaults to False.
        """
        super().__init__(
            cache_dir=cache_dir,
            dialect="mariadb",
            driver="mysqlconnector",
            user="guest",
            password="ctu-relational",
            host="relational.fel.cvut.cz",
            port=3306,
            database=database,
            time_col_dict=time_col_dict,
            keep_original_keys=keep_original_keys,
        )


class Accidents(CTUDataset):
    val_timestamp = pd.Timestamp("2003-01-01")
    test_timestamp = pd.Timestamp("2005-01-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Accidents",
            cache_dir=cache_dir,
            time_col_dict={"nesreca": "cas_nesreca"},
            keep_original_keys=False,
        )


class Airline(CTUDataset):
    val_timestamp = pd.Timestamp("2016-01-18")
    test_timestamp = pd.Timestamp("2016-01-25")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Airline",
            cache_dir=cache_dir,
            time_col_dict={"On_Time_On_Time_Performance_2016_1": "FlightDate"},
            keep_original_keys=False,
        )


class Expenditures(CTUDataset):
    """
    The Consumer Expenditure Survey (CE) collects data on expenditures, income, \
    and demographics in the United States. The public-use microdata (PUMD) files \
    provide this information for individual respondents without any information that \
    could identify respondents. PUMD files include adjustments for information that is \
    missing because respondents were unwilling or unable to provide it. The files also \
    have been adjusted to reduce the likelihood of identifying respondents, either \
    directly or through inference. The task is to predict, whether the expenditure
    is a gift or not. Household ids change from year to year - this is a property of the \
    data source.
    
    Original source: www.bls.gov
    """

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "ConsumerExpenditures",
            cache_dir=cache_dir,
            keep_original_keys=False,
        )


class Employee(CTUDataset):
    val_timestamp = pd.Timestamp("1998-02-01")
    test_timestamp = pd.Timestamp("2000-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "employee",
            cache_dir=cache_dir,
            time_col_dict={
                "dept_emp": "from_date",
                "dept_manager": "from_date",
                "employees": "hire_date",
                "salaries": "from_date",
                "titles": "from_date",
            },
            keep_original_keys=True,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Drop redundant key columns
        db.table_dict["titles"].df.drop(columns=["emp_no"], inplace=True)
        db.table_dict["dept_manager"].df.drop(columns=["emp_no", "dept_no"], inplace=True)
        db.table_dict["departments"].df.drop(columns=["dept_no"], inplace=True)
        db.table_dict["salaries"].df.drop(columns=["emp_no"], inplace=True)
        db.table_dict["dept_emp"].df.drop(columns=["emp_no", "dept_no"], inplace=True)
        db.table_dict["employees"].df.drop(columns=["emp_no"], inplace=True)

        # Convert date columns to datetime and handle NaT values
        db.table_dict["titles"].df["to_date"] = pd.to_datetime(
            db.table_dict["titles"].df["to_date"], errors="coerce"
        )
        db.table_dict["dept_manager"].df["to_date"] = pd.to_datetime(
            db.table_dict["dept_manager"].df["to_date"], errors="coerce"
        )
        db.table_dict["salaries"].df["to_date"] = pd.to_datetime(
            db.table_dict["salaries"].df["to_date"], errors="coerce"
        )
        db.table_dict["dept_emp"].df["to_date"] = pd.to_datetime(
            db.table_dict["dept_emp"].df["to_date"], errors="coerce"
        )

        return db


class LegalActs(CTUDataset):
    """
    Bulgarian court decision metadata.
    """

    val_timestamp = pd.Timestamp("2011-08-01")
    test_timestamp = pd.Timestamp("2012-03-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "legalActs",
            cache_dir=cache_dir,
            time_col_dict={
                "legalacts": "StartDate",
            },
            keep_original_keys=False,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Remove scrape fix table
        db.table_dict.pop("scrapefix")

        return db


# class Sales(CTUDataset):
#     def __init__(self, cache_dir: Optional[str] = None):
#         super().__init__(
#             "SalesDB",
#             cache_dir=cache_dir,
#             keep_original_keys=False,
#         )


class Seznam(CTUDataset):
    val_timestamp = pd.Timestamp("2014-12-01")
    test_timestamp = pd.Timestamp("2015-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Seznam",
            cache_dir=cache_dir,
            time_col_dict={
                "dobito": "month_year_datum_transakce",
                "probehnuto": "month_year_datum_transakce",
                "probehnuto_mimo_penezenku": "Month/Year",
            },
            keep_original_keys=False,
        )


class TPCC(CTUDataset):
    val_timestamp = pd.Timestamp("1996-01-01")
    test_timestamp = pd.Timestamp("1997-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcc",
            cache_dir=cache_dir,
            time_col_dict={
                "dss_lineitem": "l_shipdate",
                "dss_order": "o_orderdate",
            },
            keep_original_keys=True,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        return db


class TPCD(CTUDataset):
    val_timestamp = pd.Timestamp("1996-01-01")
    test_timestamp = pd.Timestamp("1997-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcd",
            cache_dir=cache_dir,
            time_col_dict={
                "dss_lineitem": "l_shipdate",
                "dss_order": "o_orderdate",
            },
            keep_original_keys=True,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Add missing foreign keys
        fk_col, fk_name = self._reindex_fk(
            {tn: t.df for tn, t in db.table_dict.items()},
            "dss_partsupp",
            ["ps_partkey"],
            "dss_part",
            ["p_partkey"],
        )
        db.table_dict["dss_partsupp"].fkey_col_to_pkey_table[fk_name] = "dss_part"
        db.table_dict["dss_partsupp"].df[fk_name] = fk_col

        # Drop redundant key columns
        db.table_dict["dss_lineitem"].df.drop(columns=["l_orderkey"], inplace=True)
        db.table_dict["dss_part"].df.drop(columns=["p_partkey"], inplace=True)
        db.table_dict["dss_partsupp"].df.drop(
            columns=["ps_partkey", "ps_suppkey"], inplace=True
        )
        db.table_dict["dss_region"].df.drop(columns=["r_regionkey"], inplace=True)
        db.table_dict["dss_nation"].df.drop(
            columns=["n_nationkey", "n_regionkey"], inplace=True
        )
        db.table_dict["dss_customer"].df.drop(
            columns=["c_custkey", "c_nationkey"], inplace=True
        )
        db.table_dict["dss_supplier"].df.drop(
            columns=["s_suppkey", "s_nationkey"], inplace=True
        )
        db.table_dict["dss_order"].df.drop(
            columns=["o_orderkey", "o_custkey"], inplace=True
        )

        return db


class TPCDS(CTUDataset):
    val_timestamp = pd.Timestamp("1999-05-01")
    test_timestamp = pd.Timestamp("2001-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpcds",
            cache_dir=cache_dir,
            time_col_dict={
                "store": "s_rec_start_date",
                "customer": "c_first_sales_date_sk",
                "web_page": "wp_creation_date_sk",
                "inventory": "inv_date_sk",
                "catalog_page": "cp_start_date_sk",
                "promotion": "p_start_date_sk",
                "web_site": "web_open_date_sk",
                "item": "i_rec_start_date",
                "store_returns": "sr_returned_date_sk",
                "catalog_sales": "cs_ship_date_sk",
                "call_center": "cc_open_date_sk",
                "web_sales": "ws_ship_date_sk",
            },
            keep_original_keys=False,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Replace date and time tables with simple datetime columns
        time_df = db.table_dict["time_dim"].df
        date_df = db.table_dict["date_dim"].df

        def to_time(x):
            return pd.Timedelta(
                hours=x["t_hour"], minutes=x["t_minute"], seconds=x["t_second"]
            )

        time_df["time_delta"] = db.table_dict["time_dim"].df.apply(to_time, axis=1)

        date_fk_dict = {
            tn: [
                fk for fk, fk_tn in t.fkey_col_to_pkey_table.items() if fk_tn == "date_dim"
            ]
            for tn, t in db.table_dict.items()
        }

        time_fk_dict = {
            "store_returns": "FK_time_dim_sr_return_time_sk",
            "catalog_returns": "FK_time_dim_cr_returned_time_sk",
            "web_returns": "FK_time_dim_wr_returned_time_sk",
            "store_sales": "FK_time_dim_ss_sold_time_sk",
            "catalog_sales": "FK_time_dim_cs_sold_time_sk",
            "web_sales": "FK_time_dim_ws_sold_time_sk",
        }

        for t_name, fks in date_fk_dict.items():
            for fk in fks:
                print(db.table_dict[t_name].df)
                print(date_df)
                db.table_dict[t_name].df[fk.removeprefix(f"FK_date_dim_")] = db.table_dict[
                    t_name
                ].df.join(date_df["d_date"], on=fk, how="left")["d_date"]
                db.table_dict[t_name].df.drop(columns=[fk], inplace=True)
                db.table_dict[t_name].fkey_col_to_pkey_table.pop(fk)

        for t_name, fk in time_fk_dict.items():
            db.table_dict[t_name].df[
                fk.removeprefix(f"FK_time_dim_").replace("time", "date")
            ] = db.table_dict[t_name].df.join(time_df["time_delta"], on=fk, how="left")[
                "time_delta"
            ]
            db.table_dict[t_name].df.drop(columns=[fk], inplace=True)
            db.table_dict[t_name].fkey_col_to_pkey_table.pop(fk)

        db.table_dict.pop("time_dim")
        db.table_dict.pop("date_dim")

        return db


class TPCH(CTUDataset):
    val_timestamp = pd.Timestamp("1996-01-01")
    test_timestamp = pd.Timestamp("1997-05-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "tpch",
            cache_dir=cache_dir,
            time_col_dict={"lineitem": "l_shipdate", "orders": "o_orderdate"},
            keep_original_keys=True,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Drop redundant key columns
        db.table_dict["lineitem"].df.drop(
            columns=["l_orderkey", "l_partkey", "l_suppkey"], inplace=True
        )
        db.table_dict["part"].df.drop(columns=["p_partkey"], inplace=True)
        db.table_dict["partsupp"].df.drop(
            columns=["ps_partkey", "ps_suppkey"], inplace=True
        )
        db.table_dict["region"].df.drop(columns=["r_regionkey"], inplace=True)
        db.table_dict["nation"].df.drop(
            columns=["n_nationkey", "n_regionkey"], inplace=True
        )
        db.table_dict["customer"].df.drop(
            columns=["c_custkey", "c_nationkey"], inplace=True
        )
        db.table_dict["supplier"].df.drop(
            columns=["s_suppkey", "s_nationkey"], inplace=True
        )
        db.table_dict["orders"].df.drop(columns=["o_orderkey", "o_custkey"], inplace=True)

        return db


class VOC(CTUDataset):
    # Offset to avoid issues with pandas datetime
    YEAR_OFFSET = 100

    val_timestamp = pd.Timestamp("1837-01-01")
    test_timestamp = pd.Timestamp("1867-01-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "voc",
            cache_dir=cache_dir,
            time_col_dict={"voyages": "departure_date"},
            keep_original_keys=False,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        voyages_df = db.table_dict["voyages"].df

        def conv(x: Union[datetime.date, None]):
            if x is None:
                return pd.NaT
            # Ignore dates before 1678 to avoid issues with pandas datetime
            if x.year + self.YEAR_OFFSET < 1678:
                return pd.NaT
            # Add offset to year to avoid issues with pandas datetime
            return pd.Timestamp(year=x.year + self.YEAR_OFFSET, month=x.month, day=x.day)

        voyages_df["departure_date"] = voyages_df["departure_date"].apply(conv)
        voyages_df["arrival_date"] = voyages_df["arrival_date"].apply(conv)
        voyages_df["cape_departure"] = voyages_df["cape_departure"].apply(conv)
        voyages_df["cape_arrival"] = voyages_df["cape_arrival"].apply(conv)

        db.table_dict["voyages"].df = voyages_df

        return db


class Walmart(CTUDataset):
    TIME_ORIGIN = pd.Timestamp("1970-01-01")

    val_timestamp = pd.Timestamp("2013-09-01")
    test_timestamp = pd.Timestamp("2014-04-01")

    def __init__(self, cache_dir: Optional[str] = None):
        super().__init__(
            "Walmart",
            cache_dir=cache_dir,
            time_col_dict={"weather": "date"},
            keep_original_keys=True,
        )

    def make_db(self) -> Database:
        db = super().make_db()

        # Convert sunrise and sunset to datetime
        weather_df = db.table_dict["weather"].df
        weather_df["sunrise"] = pd.to_datetime(self.TIME_ORIGIN + weather_df["sunrise"])
        weather_df["sunset"] = pd.to_datetime(self.TIME_ORIGIN + weather_df["sunset"])
        db.table_dict["weather"].df = weather_df

        return db



from relbench.datasets import register_dataset

from .db_dataset import *
from .ctu_datasets import *

register_dataset("ctu-accidents", Accidents)
register_dataset("ctu-airline", Airline)
register_dataset("ctu-expenditures", Expenditures)
register_dataset("ctu-employee", Employee)
register_dataset("ctu-employee", LegalActs)
register_dataset("ctu-seznam", Seznam)
register_dataset("ctu-tpcc", TPCC)
register_dataset("ctu-tpcd", TPCD)
register_dataset("ctu-tpcds", TPCDS)
register_dataset("ctu-tpch", TPCH)
register_dataset("ctu-voc", VOC)
register_dataset("ctu-walmart", Walmart)



from typing import Dict, List, Optional

import pandas as pd
import sqlalchemy as sa

from tqdm.std import tqdm

from relbench.base import Dataset, Database, Table

from ctu_relational.db import DBInspector, ForeignKeyDef

__ALL__ = ["DBDataset"]


class DBDataset(Dataset):
    """
    A dataset that is created from a remote relational database.

    Attributes:
        remote_url (str): The URL for connecting to the remote database.
    """

    def __init__(
        self,
        cache_dir: Optional[str] = None,
        remote_url: Optional[str] = None,
        dialect: Optional[str] = None,
        driver: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[str] = None,
        database: Optional[str] = None,
        time_col_dict: Optional[Dict[str, str]] = None,
        keep_original_keys: bool = False,
    ):
        """Create a database dataset object.

        Args:
            cache_dir (str, optional): The directory to cache the dataset. Defaults to None.
            remote_url (str, optional): The URL for connecting to the remote database in SQLAlchemy format. \
                For more information, see https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls. \
                If not defined, all following parameters have to be specified. Defaults to None.
            dialect (str, optional): The dialect for the database connection. Defaults to None.
            driver (str, optional): The driver for the database connection. Defaults to None.
            user (str, optional): The username for the database connection. Defaults to None.
            password (str, optional): The password for the database connection. Defaults to None.
            host (str, optional): The host address of the remote database. Defaults to None.
            port (str, optional): The port number for the database connection. Defaults to None.
            database (str, optional): The name of the database. Defaults to None.
            time_col_dict (Dict[str, str], optional): A dictionary mapping table names to time columns. Defaults to None.
            keep_original_keys (bool, optional): Whether to keep original primary and foreign keys \
                after duplication during re-indexing. This is useful when the keys contain information \
                beyond just their relationship to other rows. Defaults to False.
        """

        self.remote_url = (
            remote_url
            if remote_url is not None
            else self.get_url(dialect, driver, user, password, host, port, database)
        )

        self.time_col_dict = time_col_dict if time_col_dict is not None else {}

        self.keep_original_keys = keep_original_keys

        super().__init__(cache_dir)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(remote_url={self.remote_url})"

    @classmethod
    def get_url(
        cls,
        dialect: str,
        driver: str,
        user: str,
        password: str,
        host: str,
        port: str,
        database: str,
    ) -> str:
        """
        Returns the URL for connecting to the remote database in format used by SQLAlchemy.
        For more information, see https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls.

        Args:
            dialect (str): The dialect for the database connection.
            driver (str): The driver for the database connection.
            user (str): The username for the database connection.
            password (str): The password for the database connection.
            host (str): The host address of the remote database.
            port (str): The port number for the database connection.
            database (str): The name of the database.

        Returns:
            str: The URL for connecting to the remote database.
        """
        return f"{dialect}+{driver}://{user}:{password}@{host}:{port}/{database}"

    @classmethod
    def create_remote_connection(cls, remote_url: str) -> sa.Connection:
        """
        Create a new SQLAlchemy Connection instance to the remote database.
        Don't forget to close the Connection after you are done using it!

        Args:
            remote_url (str): The URL for connecting to the remote database.
                Format is dialect+driver://username:password@host:port/database

        Returns:
            Connection: The SQLAlchemy Connection instance to the remote database.
        """
        return sa.Connection(sa.create_engine(remote_url))

    def get_scheme(self) -> Dict[str, Dict[str, sa.types.TypeEngine]]:
        """Get the type scheme of the remote database.

        Returns:
            Dict[str, Dict[str, TypeEngine]]: A dictionary mapping table names to column names and their types.
        """
        remote_con = self.create_remote_connection(self.remote_url)

        inspector = DBInspector(remote_con)

        remote_md = sa.MetaData()
        remote_md.reflect(bind=inspector.engine)

        table_names = inspector.get_tables()

        table_sql__types = {}

        for t_name in table_names:

            sql_table = sa.Table(t_name, remote_md)

            table_sql__types[t_name] = {}

            for c in sql_table.columns:
                try:
                    sql_type = type(c.type.as_generic())
                except NotImplementedError:
                    sql_type = None

                table_sql__types[t_name][c.name] = sql_type

        remote_con.close()

        return table_sql__types

    def make_db(self) -> Database:
        """
        Create a Database instance from the remote database.

        Returns:
            Database: The Database instance.
        """
        remote_con = self.create_remote_connection(self.remote_url)

        inspector = DBInspector(remote_con)

        remote_md = sa.MetaData()
        remote_md.reflect(bind=inspector.engine)

        table_names = inspector.get_tables()

        df_dict: Dict[str, pd.DataFrame] = {}
        fk_dict: Dict[str, List[ForeignKeyDef]] = {}

        for t_name in tqdm(table_names, desc="Downloading tables"):

            sql_table = sa.Table(t_name, remote_md)

            dtypes: Dict[str, str] = {}
            sql_types_dict: Dict[str, sa.types.TypeEngine] = {}

            for c in sql_table.columns:
                try:
                    sql_type = type(c.type.as_generic())
                except NotImplementedError:
                    sql_type = None

                dtype = SQL_TO_PANDAS.get(sql_type, None)
                if dtype is not None:
                    dtypes[c.name] = dtype
                    sql_types_dict[c.name] = sql_type
                else:
                    print(f"Unknown data type {c.type}")

            statement = sa.select(sql_table.columns)
            query = statement.compile(remote_con.engine)
            df = pd.read_sql_query(str(query), con=remote_con, dtype=dtypes)

            for col, sql_type in sql_types_dict.items():
                if sql_type in DATE_TYPES:
                    try:
                        df[col] = pd.to_datetime(df[col])
                    except pd.errors.OutOfBoundsDatetime:
                        pass
                        # print(f"Out of bounds datetime in {t_name}.{col}", file=sys.stderr)

            # Create index column used as artificial primary key
            df.index.name = "__PK__"
            df.reset_index(inplace=True)

            df_dict[t_name] = df

            fk_dict[t_name] = inspector.get_foreign_keys(t_name)

        table_dict: Dict[str, Table] = {}

        # Re-index keys as RelBench do not support composite keys.
        # Also this way all original columns are preserved.
        for t_name in table_names:
            fkey_col_to_pkey_table: Dict[str, str] = {}

            for fk in fk_dict[t_name]:
                fk_col, fk_name = self._reindex_fk(
                    df_dict, t_name, fk.src_columns, fk.ref_table, fk.ref_columns
                )

                fkey_col_to_pkey_table[fk_name] = fk.ref_table
                df_dict[t_name][fk_name] = fk_col

            table_dict[t_name] = Table(
                df=df_dict[t_name],
                fkey_col_to_pkey_table=fkey_col_to_pkey_table,
                pkey_col="__PK__",
                time_col=self.time_col_dict.get(t_name, None),
            )

        # Remove original primary and foreign keys
        if not self.keep_original_keys:
            for t_name in table_names:
                sql_table = sa.Table(t_name, remote_md)
                table = table_dict[t_name]
                table.df.drop(
                    # Drop primary key columns
                    columns={c.name for c in sql_table.primary_key.columns}.union(
                        # Drop foreign key columns
                        {
                            c.name
                            for fk in sql_table.foreign_key_constraints
                            for c in fk.columns
                        }
                    ),
                    inplace=True,
                )

        remote_con.close()

        return Database(table_dict)

    def _reindex_fk(
        self,
        df_dict: Dict[str, pd.DataFrame],
        src_table: str,
        src_columns: List[str],
        ref_table: str,
        ref_columns: List[str],
    ):
        fk_name = f"FK_{ref_table}_" + "_".join(src_columns)

        df_src = df_dict[src_table][src_columns]
        df_ref = df_dict[ref_table]

        fk_col = df_src.merge(
            df_ref,
            how="left",
            left_on=src_columns,
            right_on=ref_columns,
        )["__PK__"]

        return fk_col, fk_name


DATE_TYPES = (sa.types.Date, sa.types.DateTime)

SQL_TO_PANDAS = {
    sa.types.BigInteger: pd.Int64Dtype(),
    sa.types.Boolean: pd.BooleanDtype(),
    sa.types.Date: "object",
    sa.types.DateTime: "object",
    sa.types.Double: pd.Float64Dtype(),
    sa.types.Enum: pd.CategoricalDtype(),
    sa.types.Float: pd.Float64Dtype(),
    sa.types.Integer: pd.Int32Dtype(),
    sa.types.Interval: "object",
    sa.types.LargeBinary: "object",
    sa.types.Numeric: pd.Float64Dtype(),
    sa.types.SmallInteger: pd.Int16Dtype(),
    sa.types.String: "string",
    sa.types.Text: "string",
    sa.types.Time: "object",
    sa.types.Unicode: "string",
    sa.types.UnicodeText: "string",
    sa.types.Uuid: "object",
}



from attr import field, define
from typing import List

__ALL__ = ["ForeignKeyDef"]


@define
class ForeignKeyDef:
    """
    Represents one foreign key.
    """

    src_columns: List[str] = field(converter=list)
    """
    The referencing columns (in this table)
    """

    ref_table: str
    """
    The referenced table name
    """

    ref_columns: List[str] = field(converter=list)
    """
    The referenced columns (in the referenced table)
    """



from typing import Dict, List, Set, Tuple

from sqlalchemy import Connection, Engine, inspect
from sqlalchemy.types import TypeEngine

from .schema import ForeignKeyDef

__ALL__ = ["DBInspector"]


class DBInspector:
    """
    A simplified helper class that allows to retrieve select basic information about a database,
    its tables, columns, and values.
    """

    def __init__(
        self,
        connection: Connection,
    ):
        """
        Initializes a new instance of the DBInspector class.

        Args:
            connection (Connection): The database connection - instance of SQLAlchemy's `Connection` class.
        """
        self._connection = connection
        self._inspect = inspect(self._connection.engine)

    @property
    def connection(self) -> Connection:
        return self._connection

    @property
    def engine(self) -> Engine:
        return self._connection.engine

    def get_tables(self) -> Set[str]:
        """
        Retrieves the names of all tables in the database.

        Returns:
            Set[str]: A set of table names.
        """
        out = self._inspect.get_table_names()
        return set(out)

    def get_columns(self, table: str) -> Dict[str, TypeEngine]:
        """
        Retrieves the columns of a specific table in the database.

        Args:
            table (str): The name of the table.

        Returns:
            Dict[str, TypeEngine]: A dictionary mapping column names to their SQLAlchemy types.
        """
        out = {col["name"]: col["type"] for col in self._inspect.get_columns(table)}
        return out

    def get_table_column_pairs(self) -> Set[Tuple[str, str]]:
        """
        Retrieves all table-column pairs in the database.

        Returns:
            Set[Tuple[str, str]]: A set of tuples representing table-column pairs.
        """
        out = set()

        for tbl in self.get_tables():
            out |= {(tbl, col) for col in self.get_columns(tbl).keys()}

        return out

    def get_primary_key(self, table: str) -> Set[str]:
        """
        Retrieves the primary key columns of a specific table in the database.

        Args:
            table (str): The name of the table.

        Returns:
            Set[str]: A set of primary key column names.
        """
        return set(self._inspect.get_pk_constraint(table)["constrained_columns"])

    def get_foreign_keys(self, table: str) -> List[ForeignKeyDef]:
        """
        Retrieves the foreign key constraints of a specific table in the database.

        Args:
            table (str): The name of the table.

        Returns:
            List[ForeignKeyDef]: A dictionary mapping sets of constrained columns to their corresponding ForeignKeyDef objects.
        """
        return [
            ForeignKeyDef(
                src_columns=fk["constrained_columns"],
                ref_table=fk["referred_table"],
                ref_columns=fk["referred_columns"],
            )
            for fk in self._inspect.get_foreign_keys(table)
        ]
